import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

# Machine learning
import xgboost as xgb
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.feature_selection import SelectKBest, f_regression, RFE


DELETE_PERC = 90

# ELBOW METHOD
def find_stable_index(metric_values: np.array, tolerance: float)-> float:
    """
    Finds the index where a metric stabilizes in an array of values.

    Parameters:
    - metric_values: A list or array of metric values.
    - tolerance: The maximum allowed difference between successive values.

    Returns:
    - The index where the metric stabilizes or None if not found.
    """
    for i in range(1, len(metric_values)):
        if abs(metric_values[i] - metric_values[i - 1]) <= tolerance:
            return i
    return None

def select_optimun(array_s: np.array, tolerance: float) -> float:
    
    inertia_diff = [array_s[i] - array_s[i - 1] for i in range(1, len(array_s))]
    index_opt = find_stable_index(inertia_diff,tolerance=tolerance)
    return index_opt


# FEATURE SELECTORS
def q1_feature_selection(x: pd.DataFrame, y: pd.DataFrame, perc: float, stats: str = None) -> pd.DataFrame:
    """ Selects the best X% features for each dependent variable (columns on y)
        
    Args:
        x (pd.DataFrame): Independent variables dataframe
        y (pd.DataFrame): Dependent variables dataframe

    Returns:
        pd.DataFrame: Pandas dataframe containing the names of selected features and their scores
    """
    name = y.columns[0]

    fs = SelectKBest(score_func=f_regression, k='all')  # Aqui podriamos poner para que coja el 25% mejor (i.e. 75)
    fs.fit(x, y[name])

    scores_df = pd.DataFrame.from_dict({ 'Score':
        dict(zip(x.columns, [s for s in fs.scores_]))},
        orient='index').T
    all_scores = scores_df.sort_values(by='Score', ascending=False)
    num_opt_features = select_optimun(np.array(all_scores['Score']), tolerance = 0.1)
    num_perc_features = round(num_opt_features / x.shape[1] * 100, 2)
    # Get only the most relevant features
    indexes = np.where(fs.scores_ >= np.percentile(fs.scores_, perc))
    scores_df = scores_df.iloc[indexes[0]]
    scores_df = scores_df.sort_values(by='Score', ascending=False)[:num_opt_features]
    
    if stats:
        plt.figure(figsize=(20, 10))
        sns.barplot(all_scores.T, orient = 'v').set(title=f'Feature importance for {name}')
        sns.despine()
        plt.axvline(num_opt_features)
        plt.hlines(all_scores.iloc[num_opt_features], 0, 800, linestyles='dashed', colors="#4D82BC", linewidth=2)
        plt.text(810, all_scores.iloc[num_opt_features], f'{num_perc_features}%', fontsize=18, fontfamily='fantasy', color="#4D82BC")
        plt.xticks([])
        plt.ylabel('f-score')
        plt.xlabel('Features')
        plt.tight_layout()
        plt.savefig(f"./results/feature_extraction/{name.replace('/', '_')}_FI.png", transparent=True)
        
    scores_df.to_pickle(f"./results/feature_extraction/{name.replace('/', '_')}_{perc}%feat.pkl")

    return scores_df

def rfe_selection(x: pd.DataFrame, y: pd.DataFrame, perc: float, estimator = None,  **kwargs) -> pd.DataFrame:
    """ Performs recursive feature elimination (RFE) using Decission Tree as estimator. 

    Args:
        x (pd.DataFrame): Independent variables dataframe
        y (pd.DataFrame): Dependent variables dataframe
        perc (float): Percentage of features to eliminate in % (0-100)

    Returns:
        pd.DataFrame: Pandas dataframe containing the names of selected features and their scores
    """
    if estimator is None:
        estimator = DecisionTreeRegressor(random_state=kwargs['random_state'])
        
    name = y.columns[0]
    num_feats = max(1, int(x.shape[1] - (perc / 100.0 * x.shape[1])))

    rfe = RFE(estimator=estimator, n_features_to_select = num_feats)
    rfe.fit(x, y[name])

    result = pd.DataFrame(rfe.estimator_.feature_importances_, index=rfe.get_feature_names_out())
    result.to_pickle(f"./results/feature_extraction/{name.replace('/', '_')}_{perc}%feat_RFE.pkl")

    return result

def xgboost_selection(x: pd.DataFrame, y: pd.DataFrame, tolerance=0.01, **kwargs) -> None:
    xgb_model = xgb.XGBRegressor(objective='reg:squarederror', random_state=kwargs['random_state'], n_jobs=kwargs['n_jobs'])
    xgb_model.fit(x, y)

    num_opt_features = select_optimun(xgb_model.feature_importances_, tolerance = tolerance)
    results = pd.DataFrame(xgb_model.feature_importances_,
                           index=xgb_model.get_booster().feature_names,
                           columns=['Score'])
    results = results.sort_values(by='Score', ascending=False)[:num_opt_features]
    
    # Log
    if False:
        xgb.plot_importance(xgb_model)
        plt.show()

    return results

def random_forest_selector(x: pd.DataFrame, y: pd.DataFrame, tolerance=0.01, **kwargs):
    feature_names = list(x.columns)
    forest = RandomForestRegressor(n_estimators= 200, random_state=kwargs['random_state'], n_jobs=kwargs['n_jobs'])
    forest.fit(x, y)
    
    num_opt_features = select_optimun(forest.feature_importances_, tolerance = tolerance)
    results = pd.DataFrame(forest.feature_importances_,
                           index=feature_names,
                           columns=['Score'])
    
    results = results.sort_values(by='Score', ascending=False)[:num_opt_features]
    
    # Plot
    if False:
        fig, ax = plt.subplots()
        forest_importances = pd.Series(forest.feature_importances_, index=feature_names)
        forest_importances.plot.bar(ax=ax)
        ax.set_title("Feature importances using MDI")
        ax.set_ylabel("Mean decrease in impurity")
        fig.tight_layout()
        plt.show()
    
    return results
 