import numpy as np
import pandas as pd
import pickle as pkl
from typing import Dict, Tuple
from nan_analysis import missing_values_table


def read_prebio(path: str, preprocess: bool = True) -> pd.DataFrame:
    """Load PREBIO2 dataset and optionally preprocess)
    
    Args:
        path (str): Path to file
        preprocess (bool, optional): Apply the preprocessing schedule designed for PREBIO2 or not. Defaults to True.

    Returns:
        pd.DataFrame: Loaded pandas DataFrame
    """
    df = pd.read_excel(path, decimal=',', header=3)
    df = df.iloc[0:201,0:1179].copy()
    return preprocess_prebio(df) if preprocess else df


def preprocess_prebio(df: pd.DataFrame) -> pd.DataFrame:
    """Applies the preprocessing steps to PREBIO2 DF 

    Args:
        df (pd.DataFrame): The PREBIO2 dataframe loaded as pandas DF

    Returns:
        pd.DataFrame: Preprocessed dataframe
    """
    # Note that we have computed some values in the 'time (h)' column by hand. Rows 199,200,201 and 202
    df = df.drop(columns=['Code'])
    df = df.drop(columns=['Tg (ºC)']) # We're dropping because it is completely empty!
    df = df.replace(regex=True, to_replace='^invalid value encountered', value=np.nan)
    df = df.replace(regex=True, to_replace='^float division by zero', value=np.nan)
    df = df.replace(regex=True, to_replace='^min', value=np.nan)
    df = df.replace(regex=True, to_replace='^max', value=np.nan)
    df = df.replace(regex=True, to_replace='^missing intrinsic state for', value=np.nan)
    for i in range(0, 4):
        df = df.replace(regex=True, to_replace=f'^288-{552 + i}', value=np.nan)
    df['Epoxide'] = df['Epoxide'].fillna('Unknown')
    
    # We drop columns with > 70% of data missing.
    missing_table = missing_values_table(df)
    more_than_70_labels = missing_table[missing_table['% of Total Values'] > 70].index
    print(more_than_70_labels)
    print(len(more_than_70_labels))
    input()
          
    df = df.drop(columns = more_than_70_labels)

    return df


def read_imputed(path: str = './data/imputed_polycarbonates.xlsx') -> pd.DataFrame:
    return pd.read_excel(path, decimal=',', header=0, index_col=False).select_dtypes('number')


def load_split_df(path: str = './data/imputed_polycarbonates.xlsx') -> tuple:
    df = read_imputed(path).select_dtypes('number')
    dependent_cols = ['Mn (kg·mol)-1', 'Mw/Mn', 'conversion']
    return df.drop(dependent_cols, axis=1), df[dependent_cols] 


def load_reduced_df() -> Dict[str, Tuple[pd.DataFrame, pd.DataFrame]]:
    with open('./data/reduced_df.pickle', 'rb') as f:
        return pkl.load(f)